using System;
using System.Collections.Generic;
using System.Text;
using System.Data;

namespace NBearMapping
{
    public class DataRowMemberSetter : IMemberSetter
    {
        private DataRowMemberSetter() { }

        #region IMemberSetter Members

        public void SetMembers(object targetObj, Dictionary<string, ObjectMember> members)
        {
            Check.Require(targetObj != null, "targetObj could not be null.");
            Check.Require(targetObj is DataRow, "targetObj MUST be a DataRow");
            Check.Require(members != null, "members could not be null.");

            DataRow row = targetObj as DataRow;

            for (int i = 0; i < row.Table.Columns.Count; ++i)
            {
                if (members.ContainsKey(row.Table.Columns[i].ColumnName) && row.Table.Columns[i].DataType == CommonUtils.GetOriginalTypeOfNullableType(members[row.Table.Columns[i].ColumnName].Type))
                    row[i] = members[row.Table.Columns[i].ColumnName].Value == null ? DBNull.Value : members[row.Table.Columns[i].ColumnName].Value;
            }
        }

        #endregion

        private static DataRowMemberSetter singleton = new DataRowMemberSetter();

        public static DataRowMemberSetter Instance
        {
            get { return singleton; }
        }
    }
}
